local mvec_to = Vector3()
local mvec_spread_direction = Vector3()
local mvec1 = Vector3()

function RaycastWeaponBase:_fire_raycast(user_unit, from_pos, direction, dmg_mul, shoot_player, spread_mul, autohit_mul, suppr_mul, shoot_through_data)
	local result = {}
	local hit_unit = nil
	shoot_player = shoot_player or false
	self._weapon_range = self._weapon_range or 20000
	local spread = self:_get_spread(user_unit)

	mvector3.set(mvec_spread_direction, direction)

	if spread then
		mvector3.spread(mvec_spread_direction, spread * (spread_mul or 1))
	end

	local ray_distance = shoot_through_data and shoot_through_data.ray_distance or self._weapon_range

	mvector3.set(mvec_to, mvec_spread_direction)
	mvector3.multiply(mvec_to, ray_distance)
	mvector3.add(mvec_to, from_pos)

	local damage = self:_get_current_damage(dmg_mul)
	local ray_from_unit = shoot_through_data and alive(shoot_through_data.ray_from_unit) and shoot_through_data.ray_from_unit or nil
	local col_ray = (ray_from_unit or World):raycast("ray", from_pos, mvec_to, "slot_mask", self._bullet_slotmask, "ignore_unit", self._setup.ignore_units)

	if shoot_through_data and shoot_through_data.has_hit_wall then
		if not col_ray then
			return result
		end

		mvector3.set(mvec1, col_ray.ray)
		mvector3.multiply(mvec1, -5)
		mvector3.add(mvec1, col_ray.position)

		local ray_blocked = World:raycast("ray", mvec1, shoot_through_data.from, "slot_mask", self._bullet_slotmask, "ignore_unit", self._setup.ignore_units, "report")

		if ray_blocked then
			return result
		end
	end

	local autoaim, suppression_enemies = self:check_autoaim(from_pos, direction)

	if self._autoaim then
		local weight = 0.1

		if col_ray and col_ray.unit:in_slot(managers.slot:get_mask("enemies")) then
			self._autohit_current = (self._autohit_current + weight) / (1 + weight)
			damage = self:get_damage_falloff(col_ray, user_unit)
			hit_unit = self._bullet_class:on_collision(col_ray, self._unit, user_unit, damage)
		elseif autoaim then
			local autohit_chance = 1
			autohit_chance = tweak_data.weapon[self._name_id].category == WeaponTweakData.WEAPON_CATEGORY_SNP and 1 or 1 - math.clamp((self._autohit_current - self._autohit_data.MIN_RATIO) / (self._autohit_data.MAX_RATIO - self._autohit_data.MIN_RATIO), 0, 1)

			if autohit_mul then
				autohit_chance = autohit_chance * autohit_mul
			end

			if math.random() < autohit_chance then
				self._autohit_current = (self._autohit_current + weight) / (1 + weight)
				damage = self:get_damage_falloff(autoaim, user_unit)
				hit_unit = self._bullet_class:on_collision(autoaim, self._unit, user_unit, damage)
				col_ray = autoaim
			else
				self._autohit_current = self._autohit_current / (1 + weight)
			end
		elseif col_ray then
			damage = self:get_damage_falloff(col_ray, user_unit)
			hit_unit = self._bullet_class:on_collision(col_ray, self._unit, user_unit, damage)
		end

		self._shot_fired_stats_table.hit = hit_unit and true or false

		if (not shoot_through_data or hit_unit) and (not self._ammo_data or not self._ammo_data.ignore_statistic) and not self._rays then
			if ray_distance == self._weapon_range then
				self._shot_fired_stats_table.skip_bullet_count = shoot_through_data and true

				managers.statistics:shot_fired(self._shot_fired_stats_table)
			end
		else
			self._shot_fired_stats_table.skip_bullet_count = nil
		end
	elseif col_ray then
		damage = self:get_damage_falloff(col_ray, user_unit)
		hit_unit = self._bullet_class:on_collision(col_ray, self._unit, user_unit, damage)
	end

	if suppression_enemies and self._suppression then
		result.enemies_in_cone = suppression_enemies
	end

	if (col_ray and col_ray.distance > 600 or not col_ray) and alive(self._obj_fire) then
		self._obj_fire:m_position(self._trail_effect_table.position)
		mvector3.set(self._trail_effect_table.normal, mvec_spread_direction)

		local trail = World:effect_manager():spawn(self._trail_effect_table)

		if col_ray then
			World:effect_manager():set_remaining_lifetime(trail, math.clamp((col_ray.distance - 600) / 10000, 0, col_ray.distance))
		end
	end

	result.hit_enemy = hit_unit

	if self._alert_events then
		result.rays = {
			col_ray
		}
	end

	if col_ray and col_ray.unit then
		local warcry_sniper_ricochet = tweak_data.weapon[self._name_id].category == WeaponTweakData.WEAPON_CATEGORY_SNP and managers.player:upgrade_value("player", "warcry_sniper_ricochet", false) == true

		repeat
			local kills, killed = nil

			if hit_unit then
				if not self._can_shoot_through_enemy and not warcry_sniper_ricochet then
					break
				end
				
				-- TRAILS GO HERE
				if warcry_sniper_ricochet then
					local trail_effect = Idstring("effects/vanilla/weapons/wpn_sniper_trail_001")
					local idstr_trail = Idstring("trail")
					local idstr_simulator_length = Idstring("simulator_length")
					local idstr_size = Idstring("size")
					local trail_effect_table = {
						effect = trail_effect,
						position = from_pos,
						normal = direction
					}
					local trail_length = World:effect_manager():get_initial_simulator_var_vector2(trail_effect, idstr_trail, idstr_simulator_length, idstr_size)

    				local trail = World:effect_manager():spawn(trail_effect_table)

    				mvector3.set_y(trail_length, col_ray.distance)
    				World:effect_manager():set_simulator_var_vector2(trail, idstr_trail, idstr_simulator_length, idstr_size, trail_length)
				end
				-- TRAILS END HERE

				killed = hit_unit.type == "death"
				local unit_type = col_ray.unit:base() and col_ray.unit:base()._tweak_table
				local is_enemy = not CopDamage.is_civilian(unit_type)
				kills = (shoot_through_data and shoot_through_data.kills or 0) + (killed and is_enemy and 1 or 0)
			end

			self._shoot_through_data.kills = kills

			if col_ray.distance < 0.1 or ray_distance - col_ray.distance < 50 then
				break
			end

			local has_hit_wall = shoot_through_data and shoot_through_data.has_hit_wall
			local has_passed_shield = shoot_through_data and shoot_through_data.has_passed_shield
			local is_shoot_through, is_shield, is_wall = nil

			if hit_unit then
				is_shoot_through = col_ray.unit:character_damage() and col_ray.unit:character_damage():dead()
			else
				local is_world_geometry = col_ray.unit:in_slot(managers.slot:get_mask("world_geometry"))

				if is_world_geometry then
					is_shoot_through = not col_ray.body:has_ray_type(Idstring("ai_vision"))

					if not is_shoot_through then
						if has_hit_wall or not self._can_shoot_through_wall then
							break
						end

						is_wall = true
					end
				else
					if not self._can_shoot_through_shield then
						break
					end

					is_shield = col_ray.unit:in_slot(8) and alive(col_ray.unit:parent())
				end
			end

			if not hit_unit and not is_shoot_through and not is_shield and not is_wall then
				break
			end

			local ray_from_unit = (hit_unit or is_shield) and col_ray.unit

			if is_shield then
				dmg_mul = (dmg_mul or 1) * 0.25
			end

			self._shoot_through_data.has_hit_wall = has_hit_wall or is_wall
			self._shoot_through_data.has_passed_shield = has_passed_shield or is_shield
			self._shoot_through_data.ray_from_unit = ray_from_unit
			self._shoot_through_data.ray_distance = ray_distance - col_ray.distance
			local is_ricocheting = killed and warcry_sniper_ricochet
			local closest_unit = is_ricocheting and self:_get_closest_target(col_ray.position, RaycastWeaponBase.RICOCHET_DISTANCE) or nil

			if is_ricocheting and closest_unit then
				self._shoot_through_data.ray_distance = math.min(RaycastWeaponBase.RICOCHET_DISTANCE, self._shoot_through_data.ray_distance * RaycastWeaponBase.RICOCHET_FALLOFF)
				local dir_to = closest_unit:movement():m_head_pos()

				mvector3.set(self._shoot_through_data.from, col_ray.position)
				mvector3.subtract(dir_to, self._shoot_through_data.from)
				mvector3.normalize(dir_to)
				managers.game_play_central:queue_fire_raycast(Application:time() + 0.0125, self._unit, user_unit, self._shoot_through_data.from, dir_to, dmg_mul, shoot_player, 0, autohit_mul, suppr_mul, self._shoot_through_data)
			else
				print("Shoot Through : Normal")
				mvector3.set(self._shoot_through_data.from, mvec_spread_direction)
				mvector3.multiply(self._shoot_through_data.from, is_shield and 5 or 40)
				mvector3.add(self._shoot_through_data.from, col_ray.position)
				managers.game_play_central:queue_fire_raycast(Application:time() + 0.0125, self._unit, user_unit, self._shoot_through_data.from, mvec_spread_direction, dmg_mul, shoot_player, spread_mul, autohit_mul, suppr_mul, self._shoot_through_data)
			end
		until true
	end

	return result
end
